import logging
import time
import random
import datetime as dt
from typing import Iterable, Iterator

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from django.conf import settings

log = logging.getLogger(__name__)

BASE = "https://zkillboard.com/api"

def _ua() -> str:
    return getattr(settings, "AUTOSRP_USER_AGENT",
           getattr(settings, "AUTO_SRP_ZKILL_USER_AGENT", "aa-autosrp"))

def _headers() -> dict:
    return {"User-Agent": _ua(), "Accept": "application/json", "Accept-Encoding": "gzip"}

def get_session(timeout: int = 45) -> requests.Session:
    s = requests.Session()
    retry = Retry(
        total=7, connect=4, read=4,
        backoff_factor=0.7,
        status_forcelist=[429, 502, 503, 504, 520, 522, 524],
        allowed_methods=["GET"],
        raise_on_status=False,
    )
    s.mount("https://", HTTPAdapter(max_retries=retry))
    s.headers.update(_headers())
    s.request_timeout = timeout  # type: ignore[attr-defined]
    return s

def month_spans_for(start: dt.datetime, end: dt.datetime) -> Iterator[tuple[int, int]]:
    y, m = start.year, start.month
    while (y < end.year) or (y == end.year and m <= end.month):
        yield (y, m)
        m += 1
        if m == 13:
            m, y = 1, y + 1

def url_for(kill_id: int) -> str:
    return f"https://zkillboard.com/kill/{kill_id}/"

def _get_json(session: requests.Session, url: str) -> list[dict]:
    try:
        r = session.get(url, timeout=getattr(session, "request_timeout", 45))
        if r.status_code != 200:
            log.debug("zKill HTTP %s for %s", r.status_code, url)
            return []
        data = r.json()
        return data if isinstance(data, list) else []
    except Exception as e:
        log.debug("zKill error %s for %s", e, url)
        return []

def fetch_losses_alliance_month(session: requests.Session, alliance_id: int, year: int, month: int, page: int) -> list[dict]:
    urls = (
        f"{BASE}/allianceID/{alliance_id}/losses/year/{year}/month/{month}/page/{page}/",
        f"{BASE}/losses/allianceID/{alliance_id}/year/{year}/month/{month}/page/{page}/",
    )
    for u in urls:
        rows = _get_json(session, u)
        if rows:
            return rows
    return []

def fetch_losses_corp_month(session: requests.Session, corp_id: int, year: int, month: int, page: int) -> list[dict]:
    urls = (
        f"{BASE}/corporationID/{corp_id}/losses/year/{year}/month/{month}/page/{page}/",
        f"{BASE}/losses/corporationID/{corp_id}/year/{year}/month/{month}/page/{page}/",
    )
    for u in urls:
        rows = _get_json(session, u)
        if rows:
            return rows
    return []

def fetch_losses_system_month(session: requests.Session, system_id: int, year: int, month: int, page: int) -> list[dict]:
    urls = (
        f"{BASE}/systemID/{system_id}/losses/year/{year}/month/{month}/page/{page}/",
        f"{BASE}/losses/systemID/{system_id}/year/{year}/month/{month}/page/{page}/",
        f"{BASE}/systemID/{system_id}/year/{year}/month/{month}/page/{page}/",
    )
    for u in urls:
        rows = _get_json(session, u)
        if rows:
            return rows
    return []

def enumerate_compact_by_orgs(session: requests.Session, start: dt.datetime, end: dt.datetime,
                              alliance_ids: Iterable[int], corp_ids: Iterable[int],
                              delay: float = 0.4) -> Iterator[tuple[int, str]]:
    seen: set[int] = set()
    for (Y, M) in month_spans_for(start, end):
        for aid in alliance_ids or []:
            page = 1
            while True:
                rows = fetch_losses_alliance_month(session, int(aid), Y, M, page)
                if not rows:
                    break
                for r in rows:
                    km = r.get("killmail_id")
                    h = (r.get("zkb") or {}).get("hash")
                    if not km or not h or km in seen:
                        continue
                    seen.add(km)
                    yield int(km), str(h)
                page += 1
                time.sleep(delay + random.random() * 0.3)
        for cid in corp_ids or []:
            page = 1
            while True:
                rows = fetch_losses_corp_month(session, int(cid), Y, M, page)
                if not rows:
                    break
                for r in rows:
                    km = r.get("killmail_id")
                    h = (r.get("zkb") or {}).get("hash")
                    if not km or not h or km in seen:
                        continue
                    seen.add(km)
                    yield int(km), str(h)
                page += 1
                time.sleep(delay + random.random() * 0.3)

def enumerate_compact_by_systems(session: requests.Session, start: dt.datetime, end: dt.datetime,
                                 system_ids: Iterable[int], delay: float = 0.4) -> Iterator[tuple[int, str]]:
    seen: set[int] = set()
    for (Y, M) in month_spans_for(start, end):
        for sid in system_ids or []:
            page = 1
            while True:
                rows = fetch_losses_system_month(session, int(sid), Y, M, page)
                if not rows:
                    break
                for r in rows:
                    km = r.get("killmail_id")
                    h = (r.get("zkb") or {}).get("hash")
                    if not km or not h or km in seen:
                        continue
                    seen.add(km)
                    yield int(km), str(h)
                page += 1
                time.sleep(delay + random.random() * 0.3)

def enumerate_compact_union(session: requests.Session, start: dt.datetime, end: dt.datetime,
                            system_ids: Iterable[int], alliance_ids: Iterable[int], corp_ids: Iterable[int],
                            delay: float = 0.4) -> Iterator[tuple[int, str]]:
    """Union of org- and system-based enumerations; de-duped by killmail_id."""
    yielded: set[int] = set()
    # orgs first (usually fewer pages), then systems
    for km_id, h in enumerate_compact_by_orgs(session, start, end, alliance_ids, corp_ids, delay):
        if km_id in yielded:
            continue
        yielded.add(km_id)
        yield km_id, h
    for km_id, h in enumerate_compact_by_systems(session, start, end, system_ids, delay):
        if km_id in yielded:
            continue
        yielded.add(km_id)
        yield km_id, h



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
from typing import Iterable, Optional

from django.db.models import QuerySet
from django.conf import settings
from eveuniverse.models import EveType
from fittings.models import Fitting, FittingItem

FITTED_FLAGS: set[int] = set(range(11, 35)) | set(range(92, 99)) | set(range(125, 133))

IGNORE_CATEGORY_IDS: set[int] = set(
    getattr(settings, "AUTOSRP_FITCHECK_IGNORE_CATEGORY_IDS", {8, 18, 20, 5})
)

@dataclass
class FitCompareResult:
    doctrine_fit_id: Optional[int]
    mode: str
    passed: bool
    missing: list[int]
    extra: list[int]
    substitutions: dict[int, int]
    notes: str


def _remove_ignored_by_group(type_ids: list[int]) -> list[int]:
    if not type_ids:
        return type_ids
    rows = list(
        EveType.objects.filter(id__in=set(type_ids)).values(
            "id", "eve_group__name", "eve_group__eve_category_id"
        )
    )
    ignore_by_name = set(getattr(settings, "AUTOSRP_FITCHECK_IGNORE_GROUP_NAMES", {"Festival Launcher"}))
    subsystem_category_id = 32
    found = {int(r["id"]) for r in rows}
    out: list[int] = []

    for r in rows:
        tid = int(r["id"])
        gname = (r.get("eve_group__name") or "")
        cat = int(r.get("eve_group__eve_category_id") or 0)
        if (gname in ignore_by_name) or (cat in IGNORE_CATEGORY_IDS and cat != subsystem_category_id):
            continue
        out.append(tid)

    out.extend([int(t) for t in type_ids if int(t) not in found])
    return out


def _expand_items(type_ids: Iterable[int]) -> Counter:
    if not type_ids:
        return Counter()
    ids = _remove_ignored_by_group([int(x) for x in type_ids if x])
    return Counter(ids)


def _expand_items_grouped(type_ids: Iterable[int]) -> Counter:
    ids = _remove_ignored_by_group([int(x) for x in type_ids if x])
    if not ids:
        return Counter()
    groups = dict(EveType.objects.filter(id__in=set(ids)).values_list("id", "eve_group_id"))
    group_list = [int(groups.get(tid, 0) or 0) for tid in ids]
    return Counter(g for g in group_list if g)


def _flatten_km_ids(val) -> list[int]:
    if not val:
        return []
    out: list[int] = []

    if isinstance(val, dict):
        has_bands = any(k in val for k in ("high", "mid", "low", "rig", "sub"))
        if has_bands:
            for k in ("high", "mid", "low", "rig", "sub"):
                try:
                    for x in (val.get(k) or []):
                        out.append(int(x))
                except Exception:
                    continue
            return out
        try:
            for k, q in val.items():
                tid = int(k)
                qty = int(q or 0) or 1
                if qty > 0:
                    out.extend([tid] * qty)
            return out
        except Exception:
            pass

    if isinstance(val, (list, tuple, set)):
        for x in val:
            try:
                out.append(int(x))
            except Exception:
                continue
        return out

    try:
        return [int(val)]
    except Exception:
        return []


def _fitted_from_km_ids(km_fitted_type_ids: Iterable[int] | dict, *, strict: bool) -> Counter:
    ids = _flatten_km_ids(km_fitted_type_ids)
    return _expand_items(ids) if strict else _expand_items_grouped(ids)


def _fitted_from_fit(fit: Fitting, *, strict: bool) -> Counter:
    try:
        rows = fit._items_by_type()
        ids = []
        for r in rows:
            tid = int(r.get("type_id") or 0)
            qty = int(r.get("qty") or 0) or 1
            ids.extend([tid] * qty)
    except Exception:
        q: QuerySet[FittingItem] = FittingItem.objects.filter(fit=fit).values("flag", "type_id", "type_fk_id", "quantity")
        ids = []
        for it in q:
            flag_val = it.get("flag")
            include = False

            if isinstance(flag_val, int):
                include = flag_val in FITTED_FLAGS
            else:
                flag_str = str(flag_val or "").lower()
                if flag_str.startswith(("hislot", "medslot", "loslot", "rigslot", "subsystem")):
                    include = True

            if not include:
                continue

            tid_raw = it.get("type_id") or it.get("type_fk_id") or 0
            try:
                tid = int(tid_raw)
            except (TypeError, ValueError):
                tid = 0
            qty = int(it.get("quantity") or 0) or 1
            if tid:
                ids.extend([tid] * qty)

    return _expand_items(ids) if strict else _expand_items_grouped(ids)


def _band_for_flag(flag_val) -> str | None:
    if isinstance(flag_val, int):
        f = int(flag_val)
        if 27 <= f <= 34:
            return "high"
        if 19 <= f <= 26:
            return "mid"
        if 11 <= f <= 18:
            return "low"
        if 92 <= f <= 98:
            return "rig"
        if 125 <= f <= 132:
            return "sub"
        return None
    s = str(flag_val or "").lower()
    if s.startswith("hislot"):
        return "high"
    if s.startswith("medslot"):
        return "mid"
    if s.startswith("loslot"):
        return "low"
    if s.startswith("rigslot"):
        return "rig"
    if s.startswith("subsystem"):
        return "sub"
    return None


def _band_counts_from_fit(fit: Fitting) -> dict[str, int]:
    bands = {"high": 0, "mid": 0, "low": 0, "rig": 0, "sub": 0}
    q: QuerySet[FittingItem] = FittingItem.objects.filter(fit=fit).values("flag", "quantity")
    for it in q:
        band = _band_for_flag(it.get("flag"))
        if not band:
            continue
        qty = int(it.get("quantity") or 0) or 1
        bands[band] += qty
    return bands


def _band_counts_from_km(val) -> dict[str, int] | None:
    if not isinstance(val, dict):
        return None
    has_bands = any(k in val for k in ("high", "mid", "low", "rig", "sub"))
    if not has_bands:
        return None
    return {
        "high": len([x for x in (val.get("high") or [])]),
        "mid": len([x for x in (val.get("mid") or [])]),
        "low": len([x for x in (val.get("low") or [])]),
        "rig": len([x for x in (val.get("rig") or [])]),
        "sub": len([x for x in (val.get("sub") or [])]),
    }


def _diff(doctrine: Counter, actual: Counter) -> tuple[list[int], list[int]]:
    missing, extra = [], []

    for k, need in doctrine.items():
        have = actual.get(k, 0)
        category = EveType.objects.filter(id=k).values_list("eve_group__eve_category_id", flat=True).first()

        if category == 32:
            pass

        if have < need:
            missing.extend([k] * (need - have))

    for k, have in actual.items():
        need = doctrine.get(k, 0)
        category = EveType.objects.filter(id=k).values_list("eve_group__eve_category_id", flat=True).first()
        if category == 32:
            pass
        if have > need:
            extra.extend([k] * (have - need))

    return missing, extra


def _candidate_fits(doctrine_id: int, ship_type_id: int) -> list[Fitting]:
    qs: QuerySet[Fitting] = Fitting.objects.all()

    try:
        fits = list(qs.filter(doctrines__id=doctrine_id, ship_type_id=ship_type_id))
        if fits:
            return fits
    except Exception:
        pass

    try:
        fits = list(qs.filter(doctrines__id=doctrine_id))
        if fits:
            return fits
    except Exception:
        pass

    return []


def _best_fit(doctrine_id: int, ship_type_id: int, km_fitted_type_ids: Iterable[int] | dict, *, strict: bool) -> tuple[
    Optional[Fitting], dict]:
    candidates = _candidate_fits(doctrine_id, ship_type_id)
    if not candidates:
        return None, {}
    actual = _fitted_from_km_ids(km_fitted_type_ids, strict=strict)
    scores = {}
    best = None
    best_missing = None
    for f in candidates:
        ref = _fitted_from_fit(f, strict=strict)
        missing, extra = _diff(ref, actual)
        score_missing = len(missing)
        score_extra = len(extra)
        scores[f.id] = {
            ("strict_missing" if strict else "loose_missing"): score_missing,
            ("strict_extra" if strict else "loose_extra"): score_extra,
        }
        if best is None or score_missing < (best_missing or 10 ** 6):
            best = f
            best_missing = score_missing
    return best, scores


def compare(doctrine_id: int, ship_type_id: int, km_fitted_type_ids: Iterable[int] | dict, strict_mode: bool) -> dict:
    km_ids_flat = _flatten_km_ids(km_fitted_type_ids)

    fit, _scores = _best_fit(doctrine_id, ship_type_id, km_ids_flat, strict=strict_mode)
    if not fit:
        return FitCompareResult(
            doctrine_fit_id=None,
            mode="strict" if strict_mode else "loose",
            passed=False,
            missing=[],
            extra=[],
            substitutions={},
            notes="No matching doctrine fit found for this hull.",
        ).__dict__

    if strict_mode:
        ref = _fitted_from_fit(fit, strict=True)
        act = _fitted_from_km_ids(km_ids_flat, strict=True)
        missing, extra = _diff(ref, act)

        d_bands = _band_counts_from_fit(fit)
        a_bands = _band_counts_from_km(km_fitted_type_ids)
        if a_bands is not None:
            bands_to_check = ("high", "mid", "low", "rig")
            has_shortfall = any(d_bands.get(b, 0) > a_bands.get(b, 0) for b in bands_to_check)
            if not has_shortfall:
                missing = []
        passed = (len(missing) == 0 and len(extra) == 0)
        mode = "strict"
        if not passed:
            ref_loose = _fitted_from_fit(fit, strict=False)
            act_loose = _fitted_from_km_ids(km_ids_flat, strict=False)
            m2, e2 = _diff(ref_loose, act_loose)
            note = "Strict comparison failed on exact types; group-level (meta) match would pass." if len(m2) == 0 else ""
        else:
            note = ""
    else:
        ref = _fitted_from_fit(fit, strict=False)
        act = _fitted_from_km_ids(km_ids_flat, strict=False)
        missing, extra = _diff(ref, act)

        d_bands = _band_counts_from_fit(fit)
        a_bands = _band_counts_from_km(km_fitted_type_ids)
        if a_bands is not None:
            bands_to_check = ("high", "mid", "low", "rig")
            has_shortfall = any(d_bands.get(b, 0) > a_bands.get(b, 0) for b in bands_to_check)
            if not has_shortfall:
                missing = []
        passed = (len(missing) == 0)
        mode = "loose"
        note = ""

    return FitCompareResult(
        doctrine_fit_id=fit.id,
        mode=mode,
        passed=passed,
        missing=missing,
        extra=extra,
        substitutions={},
        notes=note,
    ).__dict__


def compare_with_fit(doctrine_fit_id: int, ship_type_id: int, km_fitted_type_ids: Iterable[int] | dict,
                     strict_mode: bool) -> dict:
    try:
        fit = Fitting.objects.get(id=doctrine_fit_id)
    except Fitting.DoesNotExist:
        return FitCompareResult(
            doctrine_fit_id=None,
            mode="strict" if strict_mode else "loose",
            passed=False,
            missing=[],
            extra=[],
            substitutions={},
            notes=f"Selected doctrine fit {doctrine_fit_id} not found.",
        ).__dict__

    km_ids_flat = _flatten_km_ids(km_fitted_type_ids)

    if strict_mode:
        ref = _fitted_from_fit(fit, strict=True)
        act = _fitted_from_km_ids(km_ids_flat, strict=True)
        missing, extra = _diff(ref, act)
        d_bands = _band_counts_from_fit(fit)
        a_bands = _band_counts_from_km(km_fitted_type_ids)
        if a_bands is not None:
            bands_to_check = ("high", "mid", "low", "rig")
            has_shortfall = any(d_bands.get(b, 0) > a_bands.get(b, 0) for b in bands_to_check)
            if not has_shortfall:
                missing = []
        passed = (len(missing) == 0 and len(extra) == 0)
        mode = "strict"
        note = ""
        if not passed:
            ref_loose = _fitted_from_fit(fit, strict=False)
            act_loose = _fitted_from_km_ids(km_ids_flat, strict=False)
            m2, _e2 = _diff(ref_loose, act_loose)
            if len(m2) == 0:
                note = "Strict comparison failed on exact types; group-level (meta) match would pass."
    else:
        ref = _fitted_from_fit(fit, strict=False)
        act = _fitted_from_km_ids(km_ids_flat, strict=False)
        missing, extra = _diff(ref, act)
        d_bands = _band_counts_from_fit(fit)
        a_bands = _band_counts_from_km(km_fitted_type_ids)
        if a_bands is not None:
            bands_to_check = ("high", "mid", "low", "rig")
            has_shortfall = any(d_bands.get(b, 0) > a_bands.get(b, 0) for b in bands_to_check)
            if not has_shortfall:
                missing = []
        passed = (len(missing) == 0)
        mode = "loose"
        note = ""

    return FitCompareResult(
        doctrine_fit_id=fit.id,
        mode=mode,
        passed=passed,
        missing=missing,
        extra=extra,
        substitutions={},
        notes=note,
    ).__dict__


def km_fitted_typeids(km: dict) -> list[int]:
    items = (km.get("victim") or {}).get("items") or []
    candidates = [
        it for it in items
        if int(it.get("flag", -1)) in FITTED_FLAGS
        and it.get("item_type_id")
    ]

    out: list[int] = []
    for it in candidates:
        try:
            tid = int(it["item_type_id"])
        except Exception:
            continue
        out.append(tid)

    return out



from decimal import Decimal, ROUND_CEILING

from eveuniverse.models import EveType
from ..models import AppSetting, DoctrineReward, PayoutSuggestion, PenaltyScheme
from .penalties import compute_penalty_pct
from .pricing import hull_and_fit_prices, base_reward_for

def _flatten_fitted_ids(fitted) -> list[int]:
    """Accept banded dict {"high","mid","low","rig","sub"} or flat list, return flat list[int]."""
    if not fitted:
        return []
    if isinstance(fitted, dict):
        out: list[int] = []
        for k in ("high", "mid", "low", "rig", "sub"):
            try:
                out.extend(int(x) for x in (fitted.get(k) or []))
            except Exception:
                continue
        return out
    try:
        return [int(x) for x in fitted]
    except Exception:
        return []


def _to_map(v):
    if not v:
        return {}
    if isinstance(v, dict):
        out = {}
        for k, q in v.items():
            try:
                tid = int(k)
                qty = int(q or 0) or 1
                out[tid] = out.get(tid, 0) + qty
            except Exception:
                continue
        return out
    if isinstance(v, (list, tuple, set)):
        out = {}
        for x in v:
            try:
                tid = int(x)
                out[tid] = out.get(tid, 0) + 1
            except Exception:
                continue
        return out
    try:
        return {int(v): 1}
    except Exception:
        return {}


"""Payout computation"""
def compute_and_store_payout(kill, check):
    app = AppSetting.objects.filter(active=True).first()
    base, scheme = base_reward_for(kill.submission.doctrine_id, kill.ship_type_id, app, DoctrineReward)

    if scheme is None:
        try:
            scheme = PenaltyScheme.objects.filter(is_default=True).first()
        except Exception:
            scheme = None

    has_check = bool(check)
    if has_check and isinstance(check, dict):
        check_payload = {
            "missing": check.get("missing") or [],
            "extra": check.get("extra") or [],
        }
    elif has_check:
        missing = getattr(check, "missing", []) or []
        extra = getattr(check, "extra", []) or []
        check_payload = {"missing": missing, "extra": extra}
    else:
        check_payload = {"missing": [], "extra": []}

    try:
        count_subs = bool(getattr(scheme, "count_subsystems", True)) if scheme else True
    except Exception:
        count_subs = True
    only_subs_discrepancies = False
    if has_check and not count_subs:
        try:
            miss_map = _to_map(check_payload.get("missing"))
            extra_map = _to_map(check_payload.get("extra"))

            ids = list(set(miss_map.keys()) | set(extra_map.keys()))
            cats = {}
            if ids:
                try:
                    cats = {
                        int(r["id"]): int(r["eve_group__eve_category_id"] or 0)
                        for r in EveType.objects.filter(id__in=ids).values("id", "eve_group__eve_category_id")
                    }
                except Exception:
                    cats = {}

            SUB_CAT = 32

            def _keep_non_sub(tid: int) -> bool:
                cat = int(cats.get(int(tid), -1))
                if cat == -1:
                    return False
                return cat != SUB_CAT

            miss_map_ns = {tid: q for tid, q in miss_map.items() if _keep_non_sub(int(tid))}
            extra_map_ns = {tid: q for tid, q in extra_map.items() if _keep_non_sub(int(tid))}

            only_subs_discrepancies = (len(miss_map_ns) == 0 and len(extra_map_ns) == 0) and (len(miss_map) + len(extra_map) > 0)
            check_payload = {"missing": miss_map_ns, "extra": extra_map_ns}
        except Exception:
            pass

    if (scheme is None) or (not has_check):
        penalty_pct, breakdown = Decimal("0.00"), {"wrong": 0, "capped": False}
    else:
        if (not count_subs) and only_subs_discrepancies:
            penalty_pct, breakdown = Decimal("0.00"), {"wrong": 0, "capped": False}
        else:
            penalty_pct, breakdown = compute_penalty_pct(check_payload, scheme)
            if penalty_pct == 0:
                def _count_items(val):
                    try:
                        if isinstance(val, list):
                            return len(val)
                        if isinstance(val, dict):
                            total = 0
                            for v in val.values():
                                if isinstance(v, list):
                                    total += len(v)
                                elif isinstance(v, dict):
                                    total += sum(int(x or 0) for x in v.values())
                                else:
                                    try:
                                        total += int(v)
                                    except Exception:
                                        total += 0
                            return total
                        return int(val)
                    except Exception:
                        return 0

                wrong_count = (_count_items(check_payload.get("missing")) + _count_items(check_payload.get("extra")))
                if wrong_count > 0:
                    per_wrong = getattr(scheme, "per_wrong_module_pct", Decimal("0.00")) or Decimal("0.00")
                    max_cap = getattr(scheme, "max_total_deduction_pct", Decimal("100.00")) or Decimal("100.00")
                    est = Decimal(wrong_count) * Decimal(per_wrong)
                    capped = est > max_cap
                    penalty_pct = min(est, max_cap).quantize(Decimal("0.01"))
                    breakdown = {"wrong": int(wrong_count), "capped": bool(capped)}

    fitted_flat = _flatten_fitted_ids(getattr(kill, "fitted_type_ids", []))

    hull_v, fit_v = hull_and_fit_prices(kill.ship_type_id, fitted_flat, which="sell")
    base = base or Decimal("0.00")

    valuation_basis = base if base > 0 else (hull_v + fit_v)
    suggested_base = (valuation_basis * (Decimal("100.00") - penalty_pct) / Decimal("100.00")).quantize(Decimal("0.01"))

    def _round_up_quarter_million(amount: Decimal) -> Decimal:
        if amount <= 0:
            return Decimal("0.00")
        INC = Decimal("250000")
        units = (amount / INC).to_integral_value(rounding=ROUND_CEILING)
        return (units * INC).quantize(Decimal("0.01"))

    suggested = _round_up_quarter_million(suggested_base)

    PayoutSuggestion.objects.update_or_create(
        kill=kill,
        defaults=dict(
            base_reward_isk=base,
            penalty_pct=penalty_pct,
            penalty_breakdown=breakdown,
            suggested_isk=suggested,
            hull_price_isk=hull_v,
            fit_price_isk=fit_v,
        ),
    )



"""
Some helper functions, so we don't mess up other files too much
"""

# Django
from django.contrib.auth.models import User

# Alliance Auth
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter

def get_deleted_user():
    return User.objects.get_or_create(username="deleted")[0]


def get_main_for_character(character: EveCharacter) -> EveCharacter | None:
    """
    Get the main character for a given eve character

    :param character:
    :type character:
    :return:
    :rtype:
    """

    try:
        return character.character_ownership.user.profile.main_character
    except (
        AttributeError,
        EveCharacter.character_ownership.RelatedObjectDoesNotExist,
        CharacterOwnership.user.RelatedObjectDoesNotExist,
    ):
        return None


def get_user_for_character(character: EveCharacter) -> User:
    """
    Get the user for a character

    :param character:
    :type character:
    :return:
    :rtype:
    """

    try:
        return character.character_ownership.user.profile.user
    except (
        AttributeError,
        EveCharacter.character_ownership.RelatedObjectDoesNotExist,
        CharacterOwnership.user.RelatedObjectDoesNotExist,
    ):
        return get_deleted_user()



"""Hook into Alliance Auth"""

# Django
from django.utils.translation import gettext_lazy as _

from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

from autosrp import urls

class AutoSRPMenuItem(MenuItemHook):
    """This class ensures only authorized users will see the menu entry"""
    def __init__(self):
        MenuItemHook.__init__(
            self,
            _("SRP"),
            "fas fa-cube fa-fw",
            "autosrp:user_landing",
            navactive=["autosrp:"],
        )

    def render(self, request):
        """Render the menu item"""
        if request.user.has_perm("autosrp.basic_access"):
            return MenuItemHook.render(self, request)

        return ""

@hooks.register("menu_item_hook")
def register_menu():
    """Register the menu item"""
    return AutoSRPMenuItem()

@hooks.register("url_hook")
def register_urls():
    """Register app urls"""
    return UrlHook(urls, "autosrp", r"^autosrp/")



"""
Handling Discord direct messages to a user
"""

# Standard Library
from datetime import datetime
from enum import Enum
# Django
from django.contrib.auth.models import User
from django.utils import timezone

# Alliance Auth
from allianceauth.notifications import notify

# AA SRP
from autosrp import __title__
from autosrp.app_settings import (
    DISCORDPROXY_HOST,
    DISCORDPROXY_PORT,
    DISCORDPROXY_TIMEOUT,
    aa_discordnotify_installed,
    allianceauth_discordbot_installed,
    discordproxy_installed,
)
class DiscordEmbedColor(Enum):
    """
    Discord embed colors
    """

    INFO = 0x5BC0DE
    SUCCESS = 0x5CB85C
    WARNING = 0xF0AD4E
    DANGER = 0xD9534F


DISCORD_EMBED_COLOR_MAP = {
    "info": DiscordEmbedColor.INFO.value,
    "success": DiscordEmbedColor.SUCCESS.value,
    "warning": DiscordEmbedColor.WARNING.value,
    "danger": DiscordEmbedColor.DANGER.value,
}



def _aadiscordbot_send_private_message(
    user_id: int,
    title: str,
    message: str,
    embed_message: bool = True,
    level: str = "info",
) -> None:
    """
    Try to send a PM to a user on Discord via allianceauth-discordbot

    :param user_id:
    :type user_id:
    :param title:
    :type title:
    :param message:
    :type message:
    :param embed_message:
    :type embed_message:
    :param level:
    :type level:
    :return:
    :rtype:
    """

    if allianceauth_discordbot_installed():

        # Third Party
        from aadiscordbot.tasks import send_message
        from discord import Embed

        embed = Embed(
            title=str(title),
            description=message,
            color=DISCORD_EMBED_COLOR_MAP.get(level),
            timestamp=datetime.now(),
        )

        if embed_message is True:
            send_message(user_id=user_id, embed=embed)
        else:
            send_message(user_id=user_id, message=f"**{title}**\n\n{message}")


def _discordproxy_send_private_message(
    user_id: int,
    title: str,
    message: str,
    embed_message: bool = True,
    level: str = "info",
):
    """
    Try to send a PM to a user on Discord via discordproxy
    (fall back to allianceauth-discordbot if needed)

    :param user_id:
    :type user_id:
    :param title:
    :type title:
    :param message:
    :type message:
    :param embed_message:
    :type embed_message:
    :param level:
    :type level:
    :return:
    :rtype:
    """

    # Third Party
    from discordproxy.client import DiscordClient
    from discordproxy.exceptions import DiscordProxyException

    target = f"{DISCORDPROXY_HOST}:{DISCORDPROXY_PORT}"
    client = DiscordClient(target=target, timeout=DISCORDPROXY_TIMEOUT)

    try:
        if embed_message is True:
            # Third Party
            from discordproxy.discord_api_pb2 import Embed

            footer = Embed.Footer(text=str(__title__))
            embed = Embed(
                title=str(title),
                description=message,
                color=DISCORD_EMBED_COLOR_MAP.get(level),
                timestamp=timezone.now().isoformat(),
                footer=footer,
            )

            client.create_direct_message(user_id=user_id, embed=embed)
        else:
            client.create_direct_message(
                user_id=user_id, content=f"**{title}**\n\n{message}"
            )
    except DiscordProxyException as ex:

        _aadiscordbot_send_private_message(
            user_id=user_id,
            level=level,
            title=title,
            message=message,
            embed_message=embed_message,
        )


def _resolve_discord_uid(user: User) -> int:
    try:
        from allianceauth.services.modules.discord.models import DiscordUser  # type: ignore
        uid = DiscordUser.objects.filter(user=user).values_list("uid", flat=True).first()
        if uid:
            return int(uid)
    except Exception:
        pass

    return 0

def send_user_notification(
    user: User,
    title: str,
    message: dict[str, str],
    embed_message: bool = True,
    level: str = "info",
) -> None:
    """
    Send notification to user
    This creates a notification in Auth and a PM in Discord when either
    Discordproxy, AA-Discordbot or AA Discord Notifications is installed

    :param user:
    :type user:
    :param title:
    :type title:
    :param message:
    :type message:
    :param embed_message:
    :type embed_message:
    :param level:
    :type level:
    :return:
    :rtype:
    """

    if message.get("allianceauth"):
        getattr(notify, level, notify.info)(user=user, title=title, message=message["allianceauth"])

    discord_pref_enabled = True
    try:
        setting = getattr(user, "discord_setting", None)
        if setting is not None:
            discord_pref_enabled = bool(getattr(setting, "discord_enabled", True))
    except Exception:
        discord_pref_enabled = True

    if not discord_pref_enabled:
        return


    uid = _resolve_discord_uid(user)
    if uid <= 0 or not message.get("discord"):
        return
    try:
        if discordproxy_installed():
            _discordproxy_send_private_message(
                user_id=uid,
                level=level,
                title=title,
                message=message["discord"],
                embed_message=embed_message,
            )
            return
    except Exception:
        pass

    try:
        if allianceauth_discordbot_installed():
            _aadiscordbot_send_private_message(
                user_id=uid,
                level=level,
                title=title,
                message=message["discord"],
                embed_message=embed_message,
            )
    except Exception:
        pass
